from typing import List, Optional, Type

from unfolded.map_sdk.api.base import Action
from unfolded.map_sdk.transport.base import BaseNonInteractiveTransport
from unfolded.map_sdk.types import ResponseClass


class HTMLTransport(BaseNonInteractiveTransport):
    """Transport used in a static HTML map"""

    action_list: List[Action]

    def __init__(self) -> None:
        super().__init__()
        self.action_list = []

    def send_action(
        self,
        *,
        action: Action,
        response_class: Optional[Type[ResponseClass]] = None,
    ) -> None:
        # pylint:disable=unused-argument
        self.action_list.append(action)

    def render_template(self) -> str:
        self.rendered = True
        raise NotImplementedError("Static HTML maps cannot be rendered yet.")



import sys
from concurrent.futures import Future
from typing import Any, Generic, Optional, TypeVar

from unfolded.map_sdk.api.enums import EventType

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict

ResponseClass = TypeVar("ResponseClass")


class StoredFuture(Generic[ResponseClass]):
    """Shape of object stored in UnfoldedMap.futures"""

    future: Future
    response_class: Optional[ResponseClass]

    def __init__(self, future: Future, response_class: Optional[ResponseClass]) -> None:
        self.future = future
        self.response_class = response_class


# Use total=False to mark keys as optional
class EventResponse(TypedDict, total=False):
    """Shape of object returned for event handling"""

    type: Literal["event"]
    eventType: EventType
    data: Any


class MessageResponse(TypedDict, total=False):
    """Shape of object returned for messages"""

    type: Literal["response"]
    messageId: str
    data: Any


class ErrorResponse(TypedDict):
    """Shape of returned errors"""

    type: Literal["error"]
    messageId: str
    error: str



from unfolded.map_sdk.api import layer_api

DUMMY_MESSAGE_UUID = "ede3fafc-4945-41b7-ac64-16faa896a47a"
LAYER_UUID = "997c21eb-604b-49df-add9-b2fcc615a243"

ACTION_FIXTURES = [
    (
        layer_api.GetLayersAction(),
        {
            "type": "v1/map-sdk-get-layers",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [],
        },
    ),
    (
        layer_api.GetLayerByIdAction(layer_id="layer-id"),
        {
            "type": "v1/map-sdk-get-layer-by-id",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": ["layer-id"],
        },
    ),
    (
        layer_api.AddLayerAction(
            layer={
                "id": "layer-id",
                "data_id": "data-1",
                "fields": {"field-1": "value-1"},
            }
        ),
        {
            "type": "v1/map-sdk-add-layer",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                {"id": "layer-id", "dataId": "data-1", "fields": {"field-1": "value-1"}}
            ],
        },
    ),
    (
        layer_api.UpdateLayerAction(
            layer_id="layer-id", values={"fields": {"field-1": "value-1"}}
        ),
        {
            "type": "v1/map-sdk-update-layer",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                "layer-id",
                {"fields": {"field-1": "value-1"}},
            ],
        },
    ),
    (
        layer_api.RemoveLayerAction(layer_id="layer-id"),
        {
            "type": "v1/map-sdk-remove-layer",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": ["layer-id"],
        },
    ),
    (
        layer_api.GetLayerTimelineAction(),
        {
            "type": "v1/map-sdk-get-layer-timeline",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [],
        },
    ),
    (
        layer_api.UpdateLayerTimelineAction(values={"current_time": 0}),
        {
            "type": "v1/map-sdk-update-layer-timeline",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [{"currentTime": 0}],
        },
    ),
]



# test_validate_kwargs_default
import pytest

from unfolded.map_sdk.api.dataset_api import DatasetApiInteractiveMixin
from unfolded.map_sdk.errors import MapSDKException
from unfolded.map_sdk.utils.validators import validate_kwargs

from ..mocks.transport import MockInteractiveTransport


class TestValidateKwargs:
    def test_one_arg(self):

        # Default params: positional_only=None
        # pylint:disable=unused-argument
        @validate_kwargs()
        def _one_arg(foo: int, **kwargs):
            pass

        # Assert not raised as a positional arg
        _one_arg(0)

        # Assert not raised as a keyword arg
        _one_arg(foo=0)

        # Should raise with both a positional arg and a kwarg
        with pytest.raises(MapSDKException):
            _one_arg(0, kwarg="bar")

    def test_one_arg_with_default(self):
        # Default params: positional_only=None
        # pylint:disable=unused-argument
        @validate_kwargs()
        def _one_arg_with_default(foo: int = 1, **kwargs):
            pass

        # Assert not raised with no args
        # Note(Kyle): The default arg of the function is provided _after_ leaving the decorator, so
        # a default arg implies an empty `args` list.
        _one_arg_with_default()

        # Assert not raised as a positional arg
        _one_arg_with_default(0)

        # Assert not raised as a keyword arg
        _one_arg_with_default(foo=0)

        # Should raise with both a positional arg and a kwarg
        with pytest.raises(MapSDKException):
            _one_arg_with_default(0, kwarg="bar")

    def test_two_args(self):
        # pylint:disable=unused-argument
        @validate_kwargs(positional_only=None)
        def _two_args(foo: int = 1, bar: int = 2, **kwargs):
            pass

        # Assert not raised with no args
        _two_args()

        # Assert not raised as a positional arg
        _two_args(1, 2)

        # Assert not raised as a keyword arg
        _two_args(foo=1, bar=2)

        # Doesn't raise with one positional arg and one keyword arg
        _two_args(1, bar=2)

        # Should raise with both a positional arg and a kwarg
        with pytest.raises(MapSDKException):
            _two_args(1, 2, kwarg=3)

    def test_positional_only_arg(self):
        # pylint:disable=unused-argument
        @validate_kwargs(positional_only=["foo"])
        def _positional_only_arg(foo: int = 1, **kwargs):
            pass

        # Assert not raised with no args
        _positional_only_arg()

        # Assert not raised as a positional arg
        _positional_only_arg(1)

        # Assert raised as a kwarg
        with pytest.raises(MapSDKException):
            _positional_only_arg(foo=1)

        # Should raise with both a positional arg and a kwarg
        with pytest.raises(MapSDKException):
            _positional_only_arg(1, kwarg=3)

    def test_one_arg_on_class_method(self):
        class TestClass:
            # Default params are n_args=2, positional_only=None
            # pylint:disable=unused-argument
            @validate_kwargs()
            def _one_arg(self, foo: int, **kwargs):
                pass

        instance = TestClass()

        # Assert not raised as a positional arg
        instance._one_arg(0)

        # Assert not raised as a keyword arg
        instance._one_arg(foo=0)

        # Should raise with both a positional arg and a kwarg
        with pytest.raises(MapSDKException):
            instance._one_arg(0, kwarg="bar")

    def test_add_dataset(self):
        dataset_api = DatasetApiInteractiveMixin()
        dataset_api.transport = MockInteractiveTransport()

        dataset = {"type": "local", "data": "abc"}
        dataset_api.add_dataset(dataset, auto_create_layers=False)



from abc import abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import AnyHttpUrl
from pydantic import Field as PydanticField
from pydantic import StrictStr

from unfolded.map_sdk.api.base import Action, CamelCaseBaseModel, Number
from unfolded.map_sdk.api.enums import (
    ActionType,
    BasicFieldType,
    DatasetType,
    FieldType,
    TimestampFieldType,
    VectorTileAttributeType,
)
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
)
from unfolded.map_sdk.utils.validators import validate_kwargs

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import geopandas as gpd
except ImportError:
    gpd = None


RGBColor = Tuple[Number, Number, Number]
"""Red, green and blue channels of a color in [0-255] range."""


class BaseField(CamelCaseBaseModel):

    name: StrictStr
    """Unique identifier of the field."""

    label: StrictStr
    """User facing field label."""

    type: FieldType
    """Type of the field."""


class BasicField(BaseField):
    """Contains semantics around a single field in a dataset table."""

    type: BasicFieldType


class TimestampField(BaseField):
    """Contains semantics around a single timestamp field in a dataset table."""

    type: TimestampFieldType = TimestampFieldType.TIMESTAMP
    time_format: StrictStr
    """moment.js time format of this field data."""


Field = Union[BasicField, TimestampField]
"""Union of available field types."""


class BaseDataset(CamelCaseBaseModel):

    id: StrictStr
    """Unique identifier of the dataset."""

    type: DatasetType
    """Type of dataset."""

    label: StrictStr
    """Displayable dataset label."""

    color: RGBColor
    """Color label of the dataset."""

    fields: List[Field]
    """Schema describring the fields of the dataset."""


class LocalDataset(BaseDataset):
    """Dataset record representing a local dataset, with data provided by the caller."""

    type = DatasetType.LOCAL


class DatasetWithData(LocalDataset):

    data: List[List[Any]]
    """Tabular data backing the dataset, where the order of fields matches the order in `Dataset.fields`."""


class VectorTileBaseMetadata(CamelCaseBaseModel):
    """Vector tileset base metadata."""

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    data_url: AnyHttpUrl
    """URL template for tiles, with {x}/{y}/{z} placeholders"""


class _VectorAttributes(CamelCaseBaseModel):
    """Attributes of a Vector layer."""

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    attribute: StrictStr
    """The name of this attribute."""

    type: VectorTileAttributeType
    """The type of this attribute's values."""

    min: Optional[Number]
    """The minimum value of this attribute (if numeric)."""

    max: Optional[Number]
    """The maximum value of this attribute (if numeric)."""


class _VectorLayer(CamelCaseBaseModel):
    """Layer attribute of the Vector TileJSON."""

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    attributes: List[_VectorAttributes]
    """Attributes in the layer."""


class _VectorTilestats(CamelCaseBaseModel):
    """Tilestats attribute of the Vector TileJSON."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(num_layers={len(self.layers)})"

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    layers: List[_VectorLayer]
    """Layers in the tileset."""


class VectorTileEmbeddedMetadata(CamelCaseBaseModel):
    """Minimal shape for the embedded JSON metadata in a Tippecanoe vector
    tileset metadata file.
    """

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    tilestats: _VectorTilestats
    """The tilestats for this tileset."""


class VectorTileLocalMetadata(VectorTileBaseMetadata):
    """Vector tileset metadata, following the metadata format generated
    by Tippecanoe. This metadata shape can be passed to the
    map to synchronously add a vector tileset.
    see https://github.com/mapbox/tippecanoe
    """

    # JSON metadata has multiple names in different contexts
    # maxzoom and minzoom are returned by studio in camelCase
    def __init__(self, **data):
        super().__init__(
            _json=data.pop("metaJson", None) or data.pop("json", None),
            maxzoom=data.pop("maxzoom", None) or data.pop("maxZoom", None),
            minzoom=data.pop("minzoom", None) or data.pop("minZoom", None),
            **data,
        )

    class Config(VectorTileBaseMetadata.Config):
        extra = "ignore"

    metadata_url: Optional[AnyHttpUrl]
    """URL for tileset metadata."""

    bounds: Optional[Union[StrictStr, List[Number]]]
    """Tileset bounds, as an array or comma-delimited string in format "w,s,e,n"."""

    center: Optional[Union[StrictStr, List[Number]]]
    """Tileset center, as an array or comma-delimited string in format "lng,lat"."""

    maxzoom: Optional[Number]
    """Maximum zoom supported by the tileset."""

    minzoom: Optional[Number]
    """Minimum zoom supported by the tileset."""

    _json: Union[StrictStr, VectorTileEmbeddedMetadata, None] = PydanticField(
        ..., alias="json"
    )
    """Metadata for the tileset, as a JSON string or object."""


class VectorTileRemoteMetadata(VectorTileBaseMetadata):
    """Vector tileset metadata with a remote metadata URL. The remote metadata file
    should follow the metadata format generated by Tippecanoe. This metadata can
    be passed to the map to asynchronously load a vector tileset.
    see https://github.com/mapbox/tippecanoe
    """

    metadata_url: AnyHttpUrl
    """URL for tileset metadata."""


VectorTileMetadata = Union[VectorTileLocalMetadata, VectorTileRemoteMetadata]


class VectorTileDataset(BaseDataset):
    """Dataset record representing a vector tileset."""

    type = DatasetType.VECTOR_TILE

    metadata: VectorTileMetadata
    """Metadata about the vector tileset."""


class RasterTileBaseMetadata(CamelCaseBaseModel):
    """Raster tileset base metadata."""

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"


class RasterTileLocalMetadata(RasterTileBaseMetadata):
    """Raster tileset metadata in STAC Item format. STAC version must be >= 1.0.0,
    and the EO and Raster STAC extensions are required. This metadata shape can
    be passed to the map to synchronously add a raster tileset.
    see https://github.com/radiantearth/stac-spec/blob/master/item-spec/item-spec.md
    """

    metadata_url: Optional[AnyHttpUrl]
    """URL for tileset metadata."""

    type: StrictStr = "Feature"
    """Type of the raster tileset."""

    stac_version: StrictStr = PydanticField(..., alias="stac_version")
    """STAC version for tileset."""

    stac_extensions: List[StrictStr] = PydanticField(..., alias="stac_extensions")
    """A list of extensions the Item implements."""

    assets: Dict[StrictStr, Any]
    """Dictionary of asset objects that can be downloaded, each with a unique key."""


class RasterTileRemoteMetadata(RasterTileBaseMetadata):
    """Raster tileset metadata with a remote metadata URL. This metadata can
    be passed to the map to asynchronously load a raster tileset.
    """

    metadata_url: AnyHttpUrl
    """URL for tileset metadata."""


RasterTileMetadata = Union[RasterTileLocalMetadata, RasterTileRemoteMetadata]


class RasterTileDataset(BaseDataset):
    """Dataset record representing a raster tileset."""

    type = DatasetType.RASTER_TILE

    metadata: RasterTileMetadata
    """Metadata about the raster tileset."""


Dataset = Union[LocalDataset, VectorTileDataset, RasterTileDataset]


class _BaseDatasetCreationProps(CamelCaseBaseModel):

    id: Optional[StrictStr]
    """Unique identifier of the dataset."""

    type: DatasetType
    """Type of dataset."""

    label: Optional[StrictStr]
    """Displayable dataset label."""

    color: Optional[RGBColor]
    """Color label of the dataset."""


class _LocalDatasetCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a dataset with inline data."""

    type: DatasetType = PydanticField(default=DatasetType.LOCAL)

    data: Union[StrictStr, dict, List[List[Any]]]
    """Data used to create a dataset, in CSV, JSON, GeoJSON format."""


class _VectorTileDatasetCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a vector tile dataset with provided metadata."""

    type: DatasetType = DatasetType.VECTOR_TILE

    metadata: VectorTileLocalMetadata
    """Metadata about the vector tileset."""


class _VectorTileDatasetRemoteCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a vector tile dataset with remote metadata."""

    type: DatasetType = DatasetType.VECTOR_TILE

    metadata: VectorTileRemoteMetadata
    """Metadata about the vector tileset."""


class _RasterTileDatasetCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a raster tile dataset with provided metadata."""

    type: DatasetType = DatasetType.RASTER_TILE

    metadata: RasterTileLocalMetadata
    """Metadata about the raster tileset."""


class _RasterTileDatasetRemoteCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a raster tile dataset with remote metadata."""

    type: DatasetType = DatasetType.RASTER_TILE

    metadata: RasterTileRemoteMetadata
    """Metadata about the raster tileset."""


_DatasetCreationProps = Union[
    _LocalDatasetCreationProps,
    _VectorTileDatasetCreationProps,
    _RasterTileDatasetCreationProps,
]


_TileDatasetCreationProps = Union[
    _VectorTileDatasetRemoteCreationProps, _RasterTileDatasetRemoteCreationProps
]


class _DatasetUpdateProps(CamelCaseBaseModel):
    """A set of properties used to update a dataset."""

    label: Optional[StrictStr]
    """Displayable dataset label."""

    color: Optional[RGBColor]
    """Color label of the dataset."""


###########
# ACTIONS #
###########


class GetDatasetsAction(Action):
    """Action payload sent with `get_datasets` calls"""

    type = ActionType.GET_DATASETS


class GetDatasetByIdAction(Action):
    """Action payload sent with `get_dataset_by_id` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id"]

    type = ActionType.GET_DATASET_BY_ID
    dataset_id: StrictStr


class AddDatasetAction(Action):
    """Action payload sent with `add_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset"]
        options = ["auto_create_layers", "center_map"]

    type = ActionType.ADD_DATASET
    dataset: _DatasetCreationProps
    auto_create_layers: bool
    center_map: bool


class AddTileDatasetAction(Action):
    """Action payload sent with `add_tile_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset"]
        options = ["auto_create_layers", "center_map"]

    type = ActionType.ADD_TILE_DATASET
    dataset: _TileDatasetCreationProps
    auto_create_layers: bool
    center_map: bool


class UpdateDatasetAction(Action):
    """Action payload sent with `update_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id", "values"]

    type = ActionType.UPDATE_DATASET
    dataset_id: StrictStr
    values: _DatasetUpdateProps


class RemoveDatasetAction(Action):
    """Action payload sent with `remove_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id"]

    type = ActionType.REMOVE_DATASET
    dataset_id: StrictStr


class ReplaceDatasetAction(Action):
    """Action payload sent with `replace_dataset` calls"""

    class Meta(Action.Meta):
        args = ["this_dataset_id", "with_dataset"]

    type = ActionType.REPLACE_DATASET
    this_dataset_id: StrictStr
    with_dataset: _DatasetCreationProps


class GetDatasetWithDataAction(Action):
    """Action payload sent with `get_dataset_with_data` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id"]

    type = ActionType.GET_DATASET_WITH_DATA
    dataset_id: StrictStr


#########
# UTILS #
#########


def normalize_dataset(
    dataset: Union[Dataset, Dict, None],
    kwargs: Dict[str, Any],
    fields_to_exclude: Optional[List[str]] = None,
) -> Dict:
    # convert to dict to make transformations
    if isinstance(dataset, BaseDataset):
        dataset_dict = dataset.dict(exclude_none=True)
    elif dataset is not None and isinstance(dataset, dict):
        dataset_dict = dataset
    else:
        dataset_dict = kwargs

    if fields_to_exclude:
        for field in fields_to_exclude:
            if field in dataset_dict:
                del dataset_dict[field]

    return dataset_dict


def normalize_data(
    data: Union[str, dict, List[List[Any]], "pd.DataFrame", "gpd.GeoDataFrame"]
) -> Union[str, dict, List[List[Any]]]:

    if pd is not None and isinstance(data, pd.DataFrame):
        # convert to CSV string
        return data.to_csv()
    elif gpd is not None and isinstance(data, gpd.GeoDataFrame):
        # convert to GeoJSON FeatureCollection
        return data.to_json()
    else:
        return data


###########
# METHODS #
###########


class BaseDatasetApiMethods:
    @abstractmethod
    def add_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Optional[Dataset]:
        ...

    @abstractmethod
    def add_tile_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Optional[Dataset]:
        ...

    @abstractmethod
    def update_dataset(
        self, dataset_id: str, values: Union[Dataset, Dict, None] = None, **kwargs: Any
    ) -> Optional[Dataset]:
        ...

    @abstractmethod
    def remove_dataset(self, dataset_id: str) -> None:
        ...

    @abstractmethod
    def replace_dataset(
        self, this_dataset_id: str, with_dataset: Optional[dict] = None, **kwargs: Any
    ) -> Optional[Dataset]:
        ...


class BaseInteractiveDatasetApiMethods:
    @abstractmethod
    def get_datasets(self) -> List[Dataset]:
        ...

    @abstractmethod
    def get_dataset_by_id(self, dataset_id: str) -> Optional[Dataset]:
        ...

    @abstractmethod
    def get_dataset_with_data(self, dataset_id: str) -> Optional[DatasetWithData]:
        ...


class DatasetApiNonInteractiveMixin(BaseDatasetApiMethods):
    """Dataset methods that are supported in non-interactive (i.e. pure HTML) maps"""

    transport: BaseNonInteractiveTransport

    @validate_kwargs(positional_only=["dataset"])
    def add_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> None:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
                    metadata: (Dict): Object containing tileset metadata (for tiled datasets).

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
            metadata: (Dict): Object containing tileset metadata (for tiled datasets).
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields"])

        if "data" in dataset:
            dataset["data"] = normalize_data(dataset["data"])

        action = AddDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["dataset"])
    def add_tile_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> None:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    metadata: (Dict): Object containing tileset metadata.

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            metadata: (Dict): Object containing tileset metadata.
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields", "data"])

        action = AddTileDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["dataset_id", "values"])
    def update_dataset(
        self,
        dataset_id: str,
        values: Union[_DatasetUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> None:
        """Updates an existing dataset with given values.

        Args:
            dataset_id: The identifier of the dataset to update.
            values: The values to update.

        Kwargs:
            label: Displayable dataset label.
            color: Color label of the dataset.

        Returns:
            Updated dataset record.
        """

        action = UpdateDatasetAction(
            dataset_id=dataset_id, values=values if values else kwargs
        )
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["dataset_id"])
    def remove_dataset(self, dataset_id: str) -> None:
        """Removes a dataset from the map.

        Args:
            dataset_id: The identifier of the dataset to remove

        Returns:
            None
        """
        action = RemoveDatasetAction(dataset_id=dataset_id)
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["this_dataset_id", "with_dataset"])
    def replace_dataset(
        self,
        this_dataset_id: str,
        with_dataset: Union[Dataset, Dict, None] = None,
        **kwargs: Any,
    ) -> None:
        """Replaces a given dataset with a new one.

        Args:
            this_dataset_id: Identifier of the dataset to replace.
            with_dataset: Dataset details to replace the dataset with.

        Returns:
            The dataset that's now in use.
        """
        with_dataset = normalize_dataset(with_dataset, kwargs, ["fields", "data"])

        action = ReplaceDatasetAction(
            this_dataset_id=this_dataset_id,
            with_dataset=with_dataset,
        )
        self.transport.send_action(action=action, response_class=None)


class DatasetApiInteractiveMixin(
    BaseDatasetApiMethods, BaseInteractiveDatasetApiMethods
):

    transport: BaseInteractiveTransport

    def get_datasets(self) -> List[Dataset]:
        """Gets all the datasets currently available in the map.

        Returns:
            An array of datasets.
        """
        action = GetDatasetsAction()
        return self.transport.send_action_non_null(
            action=action, response_class=List[Dataset]
        )

    @validate_kwargs(positional_only=["dataset_id"])
    def get_dataset_by_id(self, dataset_id: str) -> Optional[Dataset]:
        """Retrieves a dataset by its identifier if it exists.

        Args:
            dataset_id: Identifier of the dataset to get.

        Returns:
            Dataset with a given identifier, or null if one doesn't exist.
        """
        action = GetDatasetByIdAction(dataset_id=dataset_id)
        return self.transport.send_action(action=action, response_class=Dataset)  # type: ignore

    @validate_kwargs(positional_only=["dataset"])
    def add_dataset(
        self,
        dataset: Optional[dict] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Dataset:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
                    metadata: (Dict): Object containing tileset metadata (for tiled datasets).

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
            metadata: (Dict): Object containing tileset metadata (for tiled datasets).
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields"])

        if "data" in dataset:
            dataset["data"] = normalize_data(dataset["data"])

        action = AddDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset"])
    def add_tile_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Dataset:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    metadata: (Dict): Object containing tileset metadata.

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            metadata: (Dict): Object containing tileset metadata.
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields", "data"])

        action = AddTileDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset_id", "values"])
    def update_dataset(
        self,
        dataset_id: str,
        values: Union[_DatasetUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> Dataset:
        """Updates an existing dataset with given values.

        Args:
            dataset_id: The identifier of the dataset to update.
            values: The values to update.

        Kwargs:
            label: Displayable dataset label.
            color: Color label of the dataset.

        Returns:
            Updated dataset record.
        """

        action = UpdateDatasetAction(
            dataset_id=dataset_id, values=values if values else kwargs
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset_id"])
    def remove_dataset(self, dataset_id: str) -> None:
        """Removes a dataset from the map.

        Args:
            dataset_id: The identifier of the dataset to remove

        Returns:
            None
        """
        action = RemoveDatasetAction(dataset_id=dataset_id)
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["this_dataset_id", "with_dataset"])
    def replace_dataset(
        self,
        this_dataset_id: str,
        with_dataset: Union[Dataset, Dict, None] = None,
        **kwargs: Any,
    ) -> Dataset:
        """Replaces a given dataset with a new one.

        Args:
            this_dataset_id: Identifier of the dataset to replace.
            with_dataset: Dataset details to replace the dataset with.

        Returns:
            The dataset that's now in use.
        """
        with_dataset = normalize_dataset(with_dataset, kwargs, ["fields", "data"])

        action = ReplaceDatasetAction(
            this_dataset_id=this_dataset_id,
            with_dataset=with_dataset,
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset_id"])
    def get_dataset_with_data(self, dataset_id: str) -> Optional[DatasetWithData]:
        """Retrieves a dataset record with its data for a given dataset if it exists.

        Args:
            dataset_id: Identifier of the dataset to get the record for.

        Returns:
            Dataset record with its data based on an identifier, or None if one doesn't exist.
        """
        action = GetDatasetWithDataAction(dataset_id=dataset_id)
        return self.transport.send_action(action=action, response_class=DatasetWithData)



from .dataset_api import ACTION_FIXTURES as DATASET_API_ACTION_FIXTURES
from .filter_api import ACTION_FIXTURES as FILTER_API_ACTION_FIXTURES
from .layer_api import ACTION_FIXTURES as LAYER_API_ACTION_FIXTURES
from .map_api import ACTION_FIXTURES as MAP_API_ACTION_FIXTURES

SERIALIZED_ACTIONS = [
    *DATASET_API_ACTION_FIXTURES,
    *FILTER_API_ACTION_FIXTURES,
    *LAYER_API_ACTION_FIXTURES,
    *MAP_API_ACTION_FIXTURES,
]



from enum import Enum
from typing import Dict

MESSAGE_PREFIX = "v1/"


class ActionType(str, Enum):
    """Actions that can be passed to Studio.

    Must be kept in sync with enum in
    python/modules/map-sdk/src/message-handling.ts
    """

    # Map API
    GET_VIEW = f"{MESSAGE_PREFIX}map-sdk-get-view"
    SET_VIEW = f"{MESSAGE_PREFIX}map-sdk-set-view"
    GET_VIEW_LIMITS = f"{MESSAGE_PREFIX}map-sdk-get-view-limits"
    SET_VIEW_LIMITS = f"{MESSAGE_PREFIX}map-sdk-set-view-limits"
    GET_MAP_CONTROL_VISIBILITY = f"{MESSAGE_PREFIX}map-sdk-get-map-control-visibility"
    SET_MAP_CONTROL_VISIBILITY = f"{MESSAGE_PREFIX}map-sdk-set-map-control-visibility"
    GET_SPLIT_MODE = f"{MESSAGE_PREFIX}map-sdk-get-split-mode"
    SET_SPLIT_MODE = f"{MESSAGE_PREFIX}map-sdk-set-split-mode"
    SET_THEME = f"{MESSAGE_PREFIX}map-sdk-set-theme"
    GET_MAP_CONFIG = f"{MESSAGE_PREFIX}map-sdk-get-map-config"
    SET_MAP_CONFIG = f"{MESSAGE_PREFIX}map-sdk-set-map-config"
    GET_MAP_STYLES = f"{MESSAGE_PREFIX}map-sdk-get-map-styles"

    # Filter API
    GET_FILTERS = f"{MESSAGE_PREFIX}map-sdk-get-filters"
    GET_FILTER_BY_ID = f"{MESSAGE_PREFIX}map-sdk-get-filter-by-id"
    ADD_FILTER = f"{MESSAGE_PREFIX}map-sdk-add-filter"
    UPDATE_FILTER = f"{MESSAGE_PREFIX}map-sdk-update-filter"
    REMOVE_FILTER = f"{MESSAGE_PREFIX}map-sdk-remove-filter"
    UPDATE_TIMELINE = f"{MESSAGE_PREFIX}map-sdk-update-timeline"

    # Datasets API
    GET_DATASETS = f"{MESSAGE_PREFIX}map-sdk-get-datasets"
    GET_DATASET_BY_ID = f"{MESSAGE_PREFIX}map-sdk-get-dataset-by-id"
    ADD_DATASET = f"{MESSAGE_PREFIX}map-sdk-add-dataset"
    ADD_TILE_DATASET = f"{MESSAGE_PREFIX}map-sdk-add-tile-dataset"
    UPDATE_DATASET = f"{MESSAGE_PREFIX}map-sdk-update-dataset"
    REMOVE_DATASET = f"{MESSAGE_PREFIX}map-sdk-remove-dataset"
    REPLACE_DATASET = f"{MESSAGE_PREFIX}map-sdk-replace-dataset"
    GET_DATASET_WITH_DATA = f"{MESSAGE_PREFIX}map-sdk-get-dataset-with-data"

    # Layer API
    GET_LAYERS = f"{MESSAGE_PREFIX}map-sdk-get-layers"
    GET_LAYER_BY_ID = f"{MESSAGE_PREFIX}map-sdk-get-layer-by-id"
    ADD_LAYER = f"{MESSAGE_PREFIX}map-sdk-add-layer"
    UPDATE_LAYER = f"{MESSAGE_PREFIX}map-sdk-update-layer"
    REMOVE_LAYER = f"{MESSAGE_PREFIX}map-sdk-remove-layer"
    GET_LAYER_TIMELINE = f"{MESSAGE_PREFIX}map-sdk-get-layer-timeline"
    UPDATE_LAYER_TIMELINE = f"{MESSAGE_PREFIX}map-sdk-update-layer-timeline"

    # Event API
    SET_EVENT_HANDLERS = f"{MESSAGE_PREFIX}map-sdk-set-event-handlers"
    REMOVE_EVENT_HANDLERS = f"{MESSAGE_PREFIX}map-sdk-remove-event-handlers"


class DatasetType(str, Enum):
    """Types of currently support datasets."""

    LOCAL = "local"
    VECTOR_TILE = "vector-tile"
    RASTER_TILE = "raster-tile"


class VectorTileAttributeType(str, Enum):
    """Types of Tippecanoe vector tileset attributes"""

    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"


class FilterType(str, Enum):
    """Types of currently supported filters."""

    RANGE = "range"
    SELECT = "select"
    TIME_RANGE = "time-range"
    MULTI_SELECT = "multi-select"


class FieldType(str, Enum):
    pass


class BasicFieldType(FieldType):

    BOOLEAN = "boolean"
    DATE = "date"
    GEOJSON = "geojson"
    INTEGER = "integer"
    REAL = "real"
    STRING = "string"
    ARRAY = "array"
    OBJECT = "object"
    POINT = "point"
    H3 = "h3"


class TimestampFieldType(FieldType):

    TIMESTAMP = "timestamp"


class LayerType(str, Enum):
    """Types of layers that the visualization is supported for."""

    POINT = "point"
    ARC = "arc"
    LINE = "line"
    GRID = "grid"
    HEXAGON = "hexagon"
    GEOJSON = "geojson"
    CLUSTER = "cluster"
    ICON = "icon"
    HEATMAP = "heatmap"
    H3 = "h3"
    THREE_D = "three-d"
    TRIP = "trip"
    S2 = "s2"
    RASTER_TILE = "raster-tile"
    VECTOR_TILE = "vector-tile"


class EventType(str, Enum):
    """Events that can be received from Studio.

    Must be kept in sync with enum in
    python/modules/map-sdk/src/message-handling.ts
    """

    # Map API
    ON_CLICK = f"{MESSAGE_PREFIX}map-sdk-on-click"
    ON_HOVER = f"{MESSAGE_PREFIX}map-sdk-on-hover"
    ON_VIEW_UPDATE = f"{MESSAGE_PREFIX}map-sdk-on-view-update"
    ON_GEOMETRY_SELECTION = f"{MESSAGE_PREFIX}map-sdk-on-geometry-selection"

    # Layer API
    ON_LAYER_TIMELINE_UPDATE = f"{MESSAGE_PREFIX}map-sdk-on-layer-timeline-update"

    # Filter API
    ON_FILTER_UPDATE = f"{MESSAGE_PREFIX}map-sdk-on-filter-update"

    # Extra
    ON_LOAD = f"{MESSAGE_PREFIX}map-sdk-on-load"


EVENT_HANDLER_MAP: Dict[str, EventType] = {
    "on_click": EventType.ON_CLICK,
    "on_hover": EventType.ON_HOVER,
    "on_view_update": EventType.ON_VIEW_UPDATE,
    "on_geometry_selection": EventType.ON_GEOMETRY_SELECTION,
    "on_layer_timeline_update": EventType.ON_LAYER_TIMELINE_UPDATE,
    "on_filter_update": EventType.ON_FILTER_UPDATE,
    "on_load": EventType.ON_LOAD,
}


REVERSE_EVENT_HANDLER_MAP: Dict[EventType, str] = {
    v: k for k, v in EVENT_HANDLER_MAP.items()
}


def is_event_type(event_type: str) -> bool:
    try:
        EventType(event_type)
    except ValueError:
        return False
    return True
